__author__ = 'sonneveld'

import random
import pygame
import data
import ega

MAZE_COLOUR = ega.BRIGHT_GREEN

class MazeNode(object):

    def __init__(self, id, x, y, maze_parent):
        self.id = id
        self.x = x
        self.y = y
        self.maze_parent = maze_parent

        self.colour = MAZE_COLOUR

        #start with a bunch of closed nodes
        self.up_open = False
        self.down_open = False
        self.left_open = False
        self.right_open = False

        self.is_rubble = False
        self.is_lava = False

    @property
    def x_px(self):
        return self.x*16

    @property
    def y_px(self):
        return self.y*16

    @property
    def pos_px(self):
        return self.x*16,self.y*16

    @property
    def hitbox(self):
        return pygame.Rect(self.x*16, self.y*16, 16, 16)

    @property
    def available_for_maze(self):
        return not self.is_rubble and not self.is_lava


    @property
    def open_nodes(self):
        """ nodes that have doors open to each other """
        if self.up_open: yield self.up_node
        if self.down_open: yield self.down_node
        if self.left_open: yield self.left_node
        if self.right_open: yield self.right_node

    @property
    def avail_directions(self):
        """ nodes that have doors open to each other """
        if self.up_open: yield 'up'
        if self.down_open: yield 'down'
        if self.left_open: yield 'left'
        if self.right_open: yield 'right'

    @property
    def nearby_nodes(self):
        """ nearby nodes that are cleared of rubble """
        if self.y > 0: yield self.up_node
        if self.y < self.maze_parent.height-1: yield self.down_node
        if self.x > 0: yield self.left_node
        if self.x < self.maze_parent.width-1: yield self.right_node

    def close_all(self):
        if self.up_open:
            self.up_open = False
            self.up_node.down_open = False
        if self.down_open:
            self.down_open = False
            self.down_node.up_open = False
        if self.left_open:
            self.left_open = False
            self.left_node.right_open = False
        if self.right_open:
            self.right_open = False
            self.right_node.left_open = False

    @property
    def up_node(self):
        return self.maze_parent.get_node(self.x, self.y-1)
    @property
    def down_node(self):
        return self.maze_parent.get_node(self.x, self.y+1)
    @property
    def left_node(self):
        return self.maze_parent.get_node(self.x-1, self.y)
    @property
    def right_node(self):
        return self.maze_parent.get_node(self.x+1, self.y)






class Maze(object):

    def __init__(self, width, height):
        self.width = width
        self.height = height

        self._data = None



        rubble_path = data.filepath("rubble2.png")
        rubble_image = pygame.image.load(rubble_path)
        self.rubble_image = rubble_image.convert_alpha()


        self.reset_maze()


    @property
    def width_px(self):
        return self.width*16

    @property
    def height_px(self):
        return self.height*16

    def collides_with_walls(self, rect):
        """ returns true if rect collides with any node walls  """
        rect = pygame.Rect(rect)
        maze_width_px = self.width * 16
        maze_height_px = self.height*16

        for node in self._data:
            x = node.x*16
            y = node.y*16
            if not node.up_open:
                line = pygame.Rect(x, y, 16, 0)
                if rect.colliderect(line):
                    return True
            if not node.left_open:
                line = pygame.Rect(x, y, 0, 16)
                if rect.colliderect(line):
                    return True

        # check maze edge
        line = pygame.Rect(0, maze_height_px-1, maze_width_px, 0)
        if rect.colliderect(line):
            return True
        line = pygame.Rect(maze_width_px-1, 0, 0, maze_height_px)
        if rect.colliderect(line):
            return True

        
        return False


    def collides_nodes(self, rect):
        """ yield nodes that collide with rect """

        rect = pygame.Rect(rect)

        for node in self._data:
            px_x = node.x*16
            px_y = node.y*16
            px_w = 16
            px_h = 16

            if rect.colliderect((px_x, px_y, px_w, px_h)):
                yield node



    def merge_node_ids(self, node_a, node_b):
        """ if node_a and node_b are in differnet id sets, merge them """
        if node_a.id != node_b.id:
            self.set_nodes_to_id(node_a.id, node_b.id)




    def set_ids_from_start(self, node, id):
        """ for node, if id is not None, set id then set the id for its
        open neighbours.   """
        if node.id is not None:
            return
        node.id = id

        if node.up_open:    self.set_ids_from_start(node.up_node, id)
        if node.down_open:  self.set_ids_from_start(node.down_node, id)
        if node.left_open:  self.set_ids_from_start(node.left_node, id)
        if node.right_open: self.set_ids_from_start(node.right_node, id)



    def _collect_nearby_cleared_nodes(self, node, s=None):
        """
        :type s: set
        :type node: MazeNode
        :rtype: set
        """
        if s is None:
            s = set()
        if node.is_rubble:
            return

        if node in s:
            return
        s.add(node)
        for x in node.nearby_nodes:
            self._collect_nearby_cleared_nodes(x, s)
        return s



    def _find_cleared_node_sets(self):
        """ find sets of nodes that are availble to walk through and near
        each other.  (eg, a trail of rubble could create two sets)
        :rtype: list of MazeNode
        """
        q = set(self._data)
        while q:
            n = q.pop()
            s = self._collect_nearby_cleared_nodes(n)
            if s:
                yield s
                q -= s


    def regenerate_selected(self, selected):
        """ regenerate maze for selected nodes """

        # reset ids and close selected nodes
        #id = 0
        for node in self._data:
            node.id = None
            if node in selected:
                node.close_all()
                node.is_rubble = False

        # merge ids hopefully.
        id = 0
        for node in self._data:
            if node.id is not None:
                continue
            self.set_ids_from_start(node, id)
            id += 1


        node_sets = self._find_cleared_node_sets()
        for node_set in node_sets:
            self.generate(node_set)



    def render_to_surface(self, surface):
        """ render maze walls to surface """

        surface.fill((40,40,40))

        maze_width_px = self.width * 16
        maze_height_px = self.height*16

        pygame.draw.rect(surface, MAZE_COLOUR, (0,0,maze_width_px,maze_height_px), 1)

        #grb
        for node in self._data:
            x = node.x*16
            y = node.y*16

            if node.is_rubble:
                surface.blit(self.rubble_image, (x, y))
                continue

            #if node.selected:
            #    pygame.draw.rect(surface, (100, 100, 100, 30), (x, y, 16, 16) )

            if not node.up_open:
                pygame.draw.line(surface, node.colour, (x, y), (x+15, y), 1)
            if not node.down_open:
                pygame.draw.line(surface, node.colour, (x, y+15), (x+15, y+15), 1)
            if not node.left_open:
                pygame.draw.line(surface, node.colour, (x, y), (x, y+15), 1)
            if not node.right_open:
                pygame.draw.line(surface, node.colour, (x+15, y), (x+15, y+15), 1)



    def is_incomplete(self, data):
        """ is the maze generation incomplete?
        :param data: list of maze nodes
        :return: true if not all ids are the same
        """
        prev = None
        for x in data:
            if x is None:
                return True
            if prev and x.id != prev.id:
                return True
            prev = x
        return False

    def get_node(self, x, y):
        """ get the node at x, y
        :param x: x pos
        :param y: y pos
        :rtype: MazeNode
        """
        return self._data[y*self.width + x]

    def set_nodes_to_id(self, old_id, new_id, nodes=None):
        """
        set nodes from old_id to new_id
        :param old_id: old id value
        :param new_id: new id value
        """
        if nodes is None:
            nodes = self._data
        for x in nodes:
            if x.id == old_id:
                x.id = new_id


    def reset_maze(self):
        """ reset maze with new closed nodes     """
        self._data = []
        id = 0
        for y in xrange(self.height):
            for x in xrange(self.width):
                node = MazeNode(id, x, y, self)
                id += 1
                self._data.append(node)


    def generate(self, nodes=None):
        """ simple maze generation        """

        if nodes is None:
            nodes = self._data

        while self.is_incomplete(nodes):

            # pick a random node
            x = random.randint(0, self.width-1)
            y = random.randint(0, self.height-1)
            node = self.get_node(x, y)
            if node not in nodes:
                continue
            if not node.available_for_maze:
                continue
                
            # pick random direction
            direction = random.randint(0, 3)
            if direction == 0: #UP
                if y == 0:
                    continue
                other = self.get_node(x, y-1)
                if other not in nodes:
                    continue
                if not other.available_for_maze:
                    continue
                if node.id == other.id:
                    continue
                node.up_open = True
                other.down_open = True
                self.set_nodes_to_id(other.id, node.id, nodes)

            elif direction == 1: #DOWN
                if y == self.height-1:
                    continue
                other = self.get_node(x, y+1)
                if other not in nodes:
                    continue
                if not other.available_for_maze:
                    continue
                if node.id == other.id:
                    continue
                node.down_open = True
                other.up_open = True
                self.set_nodes_to_id(other.id, node.id, nodes)


            elif direction == 2: #LEFT
                if x == 0:
                    continue
                other = self.get_node(x-1, y)
                if other not in nodes:
                    continue
                if not other.available_for_maze:
                    continue
                if node.id == other.id:
                    continue
                node.left_open = True
                other.right_open = True
                self.set_nodes_to_id(other.id, node.id, nodes)

            elif direction == 3: # RIGHT
                if x == self.width-1:
                    continue
                other = self.get_node(x+1, y)
                if other not in nodes:
                    continue
                if not other.available_for_maze:
                    continue
                if node.id == other.id:
                    continue
                node.right_open = True
                other.left_open = True
                self.set_nodes_to_id(other.id, node.id, nodes)
